#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from fiabilipy.component import Component
from fiabilipy.voter import Voter
from fiabilipy.system import System
from fiabilipy.markov import Markovprocess

__version__ = '2.7'
__all__ = ['System', 'Component', 'Voter', 'Markovprocess']



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

r""" Voter design

This modude gives tools to design voters and compute some metrics, such as the
reliability, the availability, the Mean-Time-To-Failure, and so on.

"""

from sympy import exp, Symbol, oo
from scipy.special import binom
from itertools import combinations, chain

from fiabilipy.component import Component

__all__ = ['Voter']

ALLSUBSETS = lambda n: (chain(*[combinations(range(n), ni)
                        for ni in xrange(n+1)]))

class Voter(Component):
    r""" A voter with identical components having a constant failure rate

        This class is used to describe a voter. A voter M out-of N works if
        and only if *at least* M components out of the N avaible work.

        Attributes
        ----------
        component: `Component`
            the component to be replicated by the voter
        N: int
            the initial number of components
        M: int
            the minimal number of working components
        lambda_ : float
            the constant failure rate of the voter
        mu : float, optional
            the constant maintainability rate of the voter
        initialy_avaible: boolean, optional
            whether the component is avaible at t=0 or not

        Examples
        --------
        >>> motor = Component('M', 1e-4, 3e-2)
        >>> voter = Voter(motor, 2, 3)
        >>> voter.mttf
        8333.33333333333
    """

    def __init__(self, component, M, N, lambda_=0, mu=0, initialy_avaible=True):
        name = '{} out-of {} − {}'.format(M, N, component.name)
        super(Voter, self).__init__(name=name, lambda_=lambda_, mu=mu,
                                    initialy_avaible=initialy_avaible)
        self.component = component
        self.M = M
        self.N = N

    def __repr__(self):
        return u'Voter(%s out-of %s)' % (self.M, self.N)

    def _probabilitiescomputation(self, t, method):
        """ Compute the `method` (reliability, availability, maintainability) of
            a voter, given its components, and the initial number of components
            and the minimal number of components.
        """
        prob = 0
        for k in xrange(self.M, self.N+1):
            a = getattr(self.component, method)(t)**k
            b = (1 - getattr(self.component, method)(t))**(self.N-k)
            prob += binom(self.N, k) * a * b
        return prob

    def reliability(self, t):
        r""" Compute the reliability of the voter at `t`

            This method compute the reliability of the voter at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The reliability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> voter = Voter(motor, 2, 3)
            >>> t = Symbol('t', positive=True)
            >>> voter.reliability(t)
            3.0*(-exp(-0.0001*t) + 1)*exp(-0.0002*t) + 1.0*exp(-0.0003*t)
            >>> voter.reliability(1000)
            0.974555817870510
        """
        ownrel = super(Voter, self).reliability(t)
        return ownrel * self._probabilitiescomputation(t, 'reliability')

    def maintainability(self, t):
        r""" Compute the maintainability of the voter at `t`

            This method compute the maintainability of the voter at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The maintainability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> voter = Voter(motor, 2, 3, mu=1e-3)
            >>> t = Symbol('t', positive=True)
            >>> voter.maintainability(t) #doctest: +NORMALIZE_WHITESPACE
            (1.0*(-exp(-0.03*t) + 1.0)**3 + 3.0*(-exp(-0.03*t)
                + 1.0)**2*exp(-0.03*t))*(-exp(-0.001*t) + 1.0)
            >>> voter.maintainability(1000)
            0.632120558828558
        """
        ownrel = super(Voter, self).maintainability(t)
        return ownrel * self._probabilitiescomputation(t, 'maintainability')

    def availability(self, t):
        r""" Compute the availability of the voter at `t`

            This method compute the availability of the voter at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The availability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> voter = Voter(motor, 2, 3, mu=1e-3)
            >>> t = Symbol('t', positive=True)
            >>> voter.availability(t) #doctest: +NORMALIZE_WHITESPACE
            3.0*(-0.00332225913621263*exp(-0.0301*t) +
            0.00332225913621265)*(0.00332225913621263*exp(-0.0301*t) +
            0.996677740863787)**2 + 1.0*(0.00332225913621263*exp(-0.0301*t) +
            0.996677740863787)**3
            >>> voter.availability(1000)
            0.999966961120940
        """
        ownavail = super(Voter, self).availability(t)
        return ownavail * self._probabilitiescomputation(t, 'availability')

    @property
    def mttf(self):
        r""" Compute the Mean-Time-To-Failure of the voter

            The MTTF is defined as :
                :math:`MTTF = \int_{0}^{\infty} R(t)dt`

            Returns
            -------
            out : float
                The component MTTF

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> voter = Voter(motor, 2, 3)
            >>> voter.mttf
            8333.33333333333
        """
        t = Symbol('t', positive=True)
        return self.reliability(t).integrate((t, 0, oo))

    @property
    def mttr(self):
        r""" Compute the Mean-Time-To-Repair of the voter

            The MTTR is defined as :
                :math:`MTTR = \int_{0}^{\infty} 1 - M(t)dt`

            Returns
            -------
            out : float
                The component MTTR

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> voter = Voter(motor, 2, 3, mu=1e-3)
            >>> voter.mttr
            1000.57547188695
        """
        t = Symbol('t', positive=True)
        return (1 - self.maintainability(t)).integrate((t, 0, oo))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import print_function, absolute_import
import unittest2

from random import random

from fiabilipy import System, Component, Markovprocess

class TestMarkov(unittest2.TestCase):
#The reliabily and the availability of systems is already tested. Let’s
#assume they are correct for any systems.
#The idea behind those tests is very simple. We build different systems twice :
# - one with the “standard” way, by describing a system by its reliabily block
#   diagram.
# - one with a Markov process, which is going to be tested
#once that is done, the availabilities computed have just to be compared each
#others. It a different result is found, then there is a bug (or more…)

    def setUp(self):
        #Let’s build some standard systems
        systems = {'series-parallel': System(),
                   'parallel-series': System(),
                   'simple': System()
                  }

        lambdas = {'alim':1e-4, 'motor':2e-5}
        mus = {'alim':5e-4, 'motor':2e-3}

        alim = [Component('Alim_A', lambda_=lambdas['alim'], mu=mus['alim']),
                Component('Alim_B', lambda_=lambdas['alim'], mu=mus['alim']),
               ]
        motors = [Component('Motor_A', lambda_=lambdas['motor'], mu=mus['motor']),
                  Component('Motor_B', lambda_=lambdas['motor'], mu=mus['motor']),
                 ]

        systems['simple']['E'] = alim[0]
        systems['simple'][alim[0]] = motors[0]
        systems['simple'][motors[0]] = 'S'

        systems['series-parallel']['E'] = [alim[0], alim[1]]
        systems['series-parallel'][alim[0]] = motors[0]
        systems['series-parallel'][alim[1]] = motors[1]
        systems['series-parallel'][motors[0]] = 'S'
        systems['series-parallel'][motors[1]] = 'S'

        systems['parallel-series']['E'] = [alim[0], alim[1]]
        systems['parallel-series'][alim[0]] = [motors[0], motors[1]]
        systems['parallel-series'][alim[1]] = [motors[0], motors[1]]
        systems['parallel-series'][motors[0]] = 'S'
        systems['parallel-series'][motors[1]] = 'S'

        #Let’s build the markov equivalent system
        self.components = (alim[0], alim[1], motors[0], motors[1])
        self.process = Markovprocess(self.components, {0:1}) #All the components work

        self.states = {
            'series-parallel': lambda x : (x[0]*x[2]) + (x[1]*x[3]),
            'parallel-series': lambda x : (x[0]+x[1]) * (x[2]+x[3]),
            'simple': lambda x : x[0]*x[2],
        }

        self.systems = systems

    def test_availability(self):
        #Let’s do `maxiter` checks of availability values, for times randomly
        #picked between [0, `maxtime`)
        maxiter = 1000
        maxtime = 420000
        for _ in xrange(maxiter):
            t = random() * maxtime
            for name, states in self.states.iteritems():
                self.assertAlmostEqual(self.process.value(t, states),
                                       self.systems[name].availability(t))

if __name__ == '__main__':
    unittest2.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import print_function

from numpy import zeros, binary_repr, where, array
from scipy.linalg import expm

__all__ = ['Markovprocess']

class Markovprocess(object):
    """ Initialize the markov process management of the system.

        Parameters
        ----------
        components : list
            the is the list of the components to manage
        initstates : dict or list
            describes the initial state of the process, giving the initial
            probabilities of being in each situation

        Examples
        --------
        Let S be a system of two components A and B.

        >>> from fiabilipy import Component
        >>> A, B = Component('A', 1e-2), Component('B', 1e-6)
        >>> comp = (A, B)
        >>> init = [0.8, 0.1, 0.1, 0]
        >>> process = Markovprocess(comp, init)

        * `init[0]` is the probability of having A and B working
        * `init[1]` is the probability of having A working and not B
        * `init[2]` is the probability of having B working and not A
        * `init[3]` is the probability of having neither A nor B working

        In a general way `init[i]` is the probability of having::

        >>> for c, state in enumerate(binary_repr(i, len(components))):
        >>>     if state:
        >>>         print('%s working' % components[c])
        >>>     else:
        >>>         print('%s not working' % components[c])

        As `initstates` may be very sparse, it can be given through a
        dictionnary as follow::

        >>> init = {}
        >>> init[0] = 0.8
        >>> init[1] = init[2] = 0.1
    """

    def __init__(self, components, initstates):
        self.components = tuple(components) #assert order won’t change
        self.matrix = None
        if isinstance(initstates, dict):
            N = len(self.components)
            self.initstates = array([initstates.get(x, 0)
                                     for x in xrange(2**N)])
        else:
            self.initstates = array(initstates)
        self._initmatrix()
        self._states = {}

    def _initmatrix(self):
        r""" Given a list of components, this function initialize the markov
            matrix.
        """
        N = len(self.components)
        #2^N different states
        #Let’s build the 2^(2N) matrix…
        self.matrix = zeros((2**N, 2**N))

        for i in xrange(2**N):
            currentstate = array([int(x) for x in binary_repr(i, N)])
            for j in xrange(i+1, 2**N):
                newstate = array([int(x) for x in binary_repr(j, N)])
                tocheck = where(newstate != currentstate) #Components changed
                if len(tocheck[0]) > 1: #Impossible to reach
                    continue

                component = self.components[tocheck[0][0]]#The changed component
                self.matrix[i, j] = component.lambda_
                self.matrix[j, i] = component.mu

        rowsum = self.matrix.sum(axis=1)
        self.matrix[xrange(2**N), xrange(2**N)] = -rowsum

    def _computestates(self, func):
        r""" Compute the states described by a function

            Parameters
            ----------
            func: function
                a function defining if a state is tracked or not

            Returns
            -------
            out: list
                the list of states actually tracked by `func`

            Examples
            --------
            >>> A, B, C, D = [Component(i, 1e-3) for i in xrange(4)]
            >>> comp = (A, B, C, D)
            >>> process = Markovprocess(comp, {0:1})
            >>> availablefunc = lambda x: (x[0] or x[1]) and (x[2] or x[3])
            >>> availablestates = process.computestates(states)

            This defines, for instance, the following parallel-series system::

                     | -- A -- |    | -- C -- |
                E -- |         | -- |         | -- S
                     | -- B -- |    | -- D -- |

            * `availablefunc` is the function describing when the system is
              available.
            * `availablestates` is the actual states when the system is
              available. The result is used by the :py:meth:`value` method.
        """

        N = len(self.components)
        nsquared = 2**N
        states = []
        for x in xrange(nsquared):
            s = [int(i) for i in binary_repr(nsquared - 1 - x, N)]
            if func(s):
                states.append(x)
        return states

    def value(self, t, statefunc=None):
        r""" Compute the probability of being in some states.

            Parameters
            ----------
            t : float
                when the probability must be computed
            state : function
                a function defining the state you want to know the probability

            Examples
            --------
            >>> A, B, C, D = [Component(i, 1e-3) for i in xrange(4)]
            >>> comp = (A, B, C, D)
            >>> process = Markovprocess(comp, {0:1})
            >>> availablefunc = lambda x: (x[0] or x[1]) and (x[2] or x[3])
            >>> process.value(100, statefunc=availablefunc)
            0.98197017562069511

            So, at :math:`t = 100`, the probability for the system to be
            available is approximaltly 0.982.

            If you want to know, the probability, at :math:`t = 1000` that all
            the components work but the first one, you proceed like that

            >>> allbutfirstfunc = lambda x: not x[0] and x[1] and x[2] and x[3]
            >>> allbutfirststates = process.computestates(allbutfirstfunc)
            >>> process.value(1000, states=allbutfirststates)
            0.031471429479129759
        """
        v = self.initstates.dot(expm(t*self.matrix))
        if not statefunc:
            return v
        else:
            try:
                states = self._states[statefunc]
            except KeyError:
                states = self._computestates(statefunc)
                self._states[statefunc] = states

            return v[(states, )].sum()

    def draw(self, output=None):
        r""" Print the content of the dot file needed to draw the markov process

            Parameters
            ----------
            output : file-like object, optional
                If `output` is given, then the content is written into this
                file. `output` *must* have a :py:meth:`write` method.

            Notes
            -----
            Please, see the `Graphviz <http://graphviz.org/>`_ website to have
            more information about how to transform the ouput code into a nice
            picture.
        """
        def binstr(x, N):
            """ Convert `x` to its binary reprensation over N bits

                >>> bin(2, 4)
                '0010'
            """
            return ''.join([i for i in binary_repr(x, N)])

        N = len(self.components)
        nsquared = 2**N
        data = ['digraph G {', '\trankdir=LR;']
        for i in xrange(nsquared):
            bini = binstr(nsquared - 1 - i, N)
            for j in xrange(i, nsquared):
                if not self.matrix[i, j]:
                    continue

                if i == j:
                    data.append('%s -> %s [label = "%s"]'
                                % (bini, bini, 1.0 + self.matrix[i, j]))
                else:
                    binj = binstr(nsquared - 1 - j, N)
                    data.append('%s -> %s [label = "%s"]'
                                % (bini, binj, self.matrix[i, j]))
                    data.append('%s -> %s [label = "%s"]'
                                % (binj, bini, self.matrix[j, i]))
        data.append('}')
        if not output:
            print('\n'.join(data))
        else:
            try:
                output.write('\n'.join(data) + '\n')
            except AttributeError:
                with open(output, 'w') as fobj:
                    fobj.write('\n'.join(data) + '\n')



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

r""" Reliability system design and computation

This module gives classes and functions to design complex systems and
compute some metrics, such as the reliability, the availability, the
Mean-Time-To-Failure, and so on.

"""

from numpy import empty, ones, delete
from sympy import exp, Symbol, oo
from scipy.special import binom
from itertools import combinations, chain
from collections import Iterable
import networkx as nx

from fiabilipy import Component

__all__ = ['System']

ALLSUBSETS = lambda n: (chain(*[combinations(range(n), ni)
                        for ni in xrange(n+1)]))


class System(object):
    r""" Describe a system with different components.

        The components are linked together thanks to a reliability diagram.
        This reliability diagram is represented by a graph. This graph
        *must* have two special nodes called `E` and `S`. `E` represents the
        start of the system and `S` its end (names stand for “Entrée”
        (start) and “Sortie” (end) in French).

        Examples
        --------

        Let’s have a look to the following system::

                 | -- C0 -- |
            E -- |          | -- C2 -- S
                 | -- C1 -- |

        Thus, to represent such a system, you must create the three
        components C0, C1 and C2 and link them.

        >>> C = [Component(i, 1e-4) for i in xrange(3)]
        >>> S = System()
        >>> S['E'] = [C[0], C[1]]
        >>> S[C[0]] = [C[2]]
        >>> S[C[1]] = [C[2]]
        >>> S[C[2]] = ['S']

        So, you can use the `System` object as a simple python dictionnary
        where each key is a component and the value associated it the list
        of the component’s successors.
    """

    def __init__(self, graph=None):
        self._graph = nx.DiGraph(graph)
        self._cache = {}
        self._t = Symbol('t', positive=True)

    def __getitem__(self, component):
        return self._graph[component]

    def __setitem__(self, component, successors):
        #Let’s do different checks before inserting the element
        if not isinstance(successors, Iterable):
            if not isinstance(successors, Component):
                msg = u'successors must be a list of components, a component '
                raise ValueError(msg)
            successors = [successors]
        if component != 'E' and 'E' not in self._graph:
            msg = u"'E' must be the first inserted component"
            raise ValueError(msg)
        for successor in successors:
            if successor != 'S':
                successor._systems.add(self)
            self._graph.add_edge(component, successor)

        #reset the cache
        self._cache = {}

    def __delitem__(self, component):
        for c in self._graph:
            try:
                self._graph.remove_edge(c, component)
            except nx.NetworkXError: #i.e. edge(c, component) does not exist
                pass
            except AttributeError:
                assert self._graph[c] == 'S'
        self._graph.remove_node(component)
        if component not in self.components:
            component._systems.remove(self)
        #reset the cache
        self._cache = {}

    def __len__(self):
        return len(self._graph)

    def __repr__(self):
        return u'I\'m a system'

    def copy(self):
        r""" Return a copy of the system.

            Returns
            -------
            out: System
                A copy of the current system

            Notes
            -----
                The components are the same (same reference).
                Only the internal graph is new
        """
        _copy = System()
        _copy['E'] = [] #'E' must be the first inserted component
        for c in self._graph:
            _copy[c] = self[c][:]
        return _copy

    @property
    def components(self):
        r""" The list of the components used by the system

            Returns
            -------
            out: list
                the list of the components used by the system, except `E` and
                `S`
        """
        return [comp for comp in self._graph if comp not in ('E', 'S')]

    def _probabilitiescomputation(self, t, method):
        """ Given a system and a `method` (either availability or
            maintainability or reliability), this method evaluates the asking
            value by exploring the graph at time `t`.
        """
        #TODO : improve complexity ?
        #   n
        # P(U a_i) = sum     (-1)^{-1+|s|} P(^a_i)
        #  i=1      s\in[1,n],              i\in s
        #           s != {}
        #
        paths = self.successpaths
        R = 0.0
        for S in ALLSUBSETS(len(paths)):
            if not S:
                continue
            comps = set([c for i in S for c in paths[i][1:-1]])
            r = reduce(lambda x, y:x*getattr(y, method)(t), comps, 1)
            R += -r if len(S) % 2 == 0 else r
        return R

    def availability(self, t):
        r""" Compute the availability of the whole system

            This method compute the availability of the system at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The availability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> power = Component('P', 1e-6, 2e-4)
            >>> t = Symbol('t', positive=True)
            >>> S = System()
            >>> S['E'] = [power]
            >>> S[power] = [motor]
            >>> S[motor] = 'S'
            >>> S.availability(t) #doctest: +NORMALIZE_WHITESPACE
            (200/201 + exp(-201*t/1000000)/201)*(300/301 +
            exp(-301*t/10000)/301)
            >>> S.availability(1000)
            0.995774842225189
        """
        try:
            formula = self._cache['availability']
        except KeyError:
            formula = self._probabilitiescomputation(self._t, 'availability')
            self._cache['availability'] = formula

        if isinstance(t, Symbol):
            return formula.nsimplify()
        else:
            return formula.subs(self._t, t).evalf()

    def reliability(self, t):
        r""" Compute the reliability of the whole system

            This method compute the reliability of the system at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The reliability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> power = Component('P', 1e-6, 2e-4)
            >>> t = Symbol('t', positive=True)
            >>> S = System()
            >>> S['E'] = [power]
            >>> S[power] = [motor]
            >>> S[motor] = 'S'
            >>> S.reliability(t)
            exp(-101*t/1000000)
            >>> S.reliability(1000)
            0.903933032885864
        """
        try:
            formula = self._cache['reliability']
        except KeyError:
            formula = self._probabilitiescomputation(self._t, 'reliability')
            self._cache['reliability'] = formula

        if isinstance(t, Symbol):
            return formula.nsimplify()
        else:
            return formula.subs(self._t, t).evalf()

    def maintainability(self, t):
        r""" Compute the maintainability of the whole system

            This method compute the maintainability of the system at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The maintainability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> power = Component('P', 1e-6, 2e-4)
            >>> t = Symbol('t', positive=True)
            >>> S = System()
            >>> S['E'] = [power]
            >>> S[power] = [motor]
            >>> S[motor] = 'S'
            >>> S.maintainability(t)
            (1 - exp(-3*t/100))*(1 - exp(-t/5000))
            >>> S.maintainability(1000)
            0.181269246922001
        """
        try:
            formula = self._cache['maintainability']
        except KeyError:
            formula = self._probabilitiescomputation(self._t, 'maintainability')
            self._cache['maintainability'] = formula

        if isinstance(t, Symbol):
            return formula.nsimplify()
        else:
            return formula.subs(self._t, t).evalf()

    @property
    def mttf(self):
        r""" Compute the Mean-Time-To-Failure of the system

            The MTTF is defined as :
                :math:`MTTF = \int_{0}^{\infty} R(t)dt`

            Returns
            -------
            out : float
                The system MTTF

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> power = Component('P', 1e-6, 2e-4)
            >>> S = System()
            >>> S['E'] = [power]
            >>> S[power] = [motor]
            >>> S[motor] = 'S'
            >>> S.mttf
            1000000/101
        """
        try:
            return self._cache['mttf']
        except KeyError:
            t = Symbol('t', positive=True)
            self._cache['mttf'] = self.reliability(t).integrate((t, 0, oo))
            return self._cache['mttf']

    @property
    def mttr(self):
        r""" Compute the Mean-Time-To-Repair of the system

            The MTTR is defined as :
                :math:`MTTF = \int_{0}^{\infty} 1 - M(t)dt`

            Returns
            -------
            out : float
                The system MTTR

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> power = Component('P', 1e-6, 2e-4)
            >>> S = System()
            >>> S['E'] = [power]
            >>> S[power] = [motor]
            >>> S[motor] = 'S'
            >>> S.mttr
            2265100/453
        """
        try:
            return self._cache['mttr']
        except KeyError:
            t = Symbol('t', positive=True)
            mttr = (1 - self.maintainability(t)).integrate((t, 0, oo))
            self._cache['mttr'] = mttr
            return self._cache['mttr']


    @property
    def successpaths(self):
        r""" Return all the success paths of the reliability diagram

            A success path is defined as a path from 'E' to 'S'.

            Returns
            -------
            out : list of paths
                the list of all the success paths. A path, is defined as a list
                of components

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> powers = [Component('P{}'.format(i), 1e-6, 2e-4) for i in (0,1)]
            >>> S = System()
            >>> S['E'] = [powers[0], powers[1]]
            >>> S[powers[0]] = S[powers[1]] = [motor]
            >>> S[motor] = 'S'
            >>> S.successpaths #doctest: +NORMALIZE_WHITESPACE
            [['E', Component(P0), Component(M), 'S'],
             ['E', Component(P1), Component(M), 'S']]
        """
        try:
            return self._cache['successpaths']
        except KeyError:
            self._cache['successpaths'] = list(self.findallpaths('E', 'S'))
            return self._cache['successpaths']

    def findallpaths(self, start='E', end='S'):
        r""" Find all paths between two components in the reliability diagram

            Parameters
            ----------
            start : Component, optional
                find paths from this component
            end : Component, optional
                find paths to this component

            Returns
            -------
            out : iterator
                an iterator on the paths from `start` to `stop`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> powers = [Component('P{}'.format(i), 1e-6, 2e-4) for i in (0,1)]
            >>> S = System()
            >>> S['E'] = [powers[0], powers[1]]
            >>> S[powers[0]] = S[powers[1]] = [motor]
            >>> S[motor] = 'S'
            >>> list(S.findallpaths(start=powers[0])) #doctest: +NORMALIZE_WHITESPACE
            [[Component(P0), Component(M), 'S']]
        """
        return nx.all_simple_paths(self._graph, start, end)

    def minimalcuts(self, order=1):
        r""" List the minimal cuts of the system of order <= `order`

            A minimal cut of order :math:`n`, is a set of :math:`n` components,
            such as if there all unavailable, the whole system is unavailable.

            This function aims to find out every minimal cuts of order inferior
            to `order`.

            Parameters
            ----------
            order : int, optional
                The maximal order to look for.

            Returns
            -------
            out : list of frozensets
                each frozenset contains the components that constitute a minimal
                cut

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> powers = [Component('P{}'.format(i), 1e-6, 2e-4) for i in (0,1)]
            >>> S = System()
            >>> S['E'] = [powers[0], powers[1]]
            >>> S[powers[0]] = S[powers[1]] = [motor]
            >>> S[motor] = 'S'
            >>> S.minimalcuts(order=1) #doctest: +ELLIPSIS
            [frozenset(...)]
            >>> S.minimalcuts(order=2) #doctest: +ELLIPSIS
            [frozenset(...), frozenset(...)]
        """
        paths = self.successpaths
        incidence = empty((len(paths), len(self.components)))

        for path in xrange(len(paths)):
            for comp in xrange(len(self.components)):
                if self.components[comp] in paths[path]:
                    incidence[path, comp] = 1
                else:
                    incidence[path, comp] = 0

        pairs = list(self.components)
        minimal = []

        for k in xrange(1, order+1):
            if incidence.shape[1] == 0: #No more minimalcuts
                break
            #Let’s looking for column of ones
            vones = ones(len(paths))
            indicetodelete = []
            for comp in xrange(len(pairs)):
                if (incidence[:, comp] == vones).all():
                    if isinstance(pairs[comp], frozenset):
                        minimal.append(pairs[comp])
                    else:
                        minimal.append(frozenset([pairs[comp]]))
                    indicetodelete.append(comp)

            if k >= order:
                #so it’s useless to compute newpairs and the new incidence
                #matrix because they won’t be used anymore.
                continue

            incidence = delete(incidence, indicetodelete, axis=1)
            pairs = [p for i, p in enumerate(pairs) if i not in indicetodelete]
            newpairs = list(combinations(range(len(pairs)), k+1))
            incidence_ = empty((len(paths), len(newpairs)))
            for x in xrange(incidence_.shape[0]):
                for y in xrange(incidence_.shape[1]):
                    value = 0
                    for comp in newpairs[y]:
                        if incidence[x, comp]:
                            value = 1
                            break
                    incidence_[x, y] = value

            incidence = incidence_
            pairs = [frozenset([pairs[x] for x in p]) for p in newpairs]

        return minimal

    def faulttreeanalysis(self, output=None, order=2):
        r""" Build the fault tree analysis of the system

            Print (or write) the content of the dot file needed to draw the
            fault tree of the system.

            Parameters
            ----------
            output : file-like object, optional
                If `output` is given, then the content is written into this
                file. `output` *must* have a :py:meth:`write` method.

            order : int, optional
                This is the maximum order of the minimal cuts the function looks
                for.

            Notes
            -----
            Please, see the `Graphviz <http://graphviz.org/>` website for more
            information about how to transform the ouput code into a nice
            picture.

        """
        #TODO the tree needs to be simplified
        cuts = self.minimalcuts(order)
        data = ['digraph G {']
        data.append('\t"not_S" -> "or"')
        for i, cut in enumerate(cuts):
            data.append('\tor -> and_%s' % i)
            for comp in cut:
                data.append('\tand_%s -> "%s"' % (i, comp.name))
        data.append('}')

        if not output:
            print('\n'.join(data))
        else:
            try:
                output.write('\n'.join(data) + '\n')
            except AttributeError:
                with open(output, 'w') as fobj:
                    fobj.write('\n'.join(data) + '\n')


    def draw(self):
        r""" Draw the system

            Draw the system with graphviz.

            Examples
            --------
            >>> import pylab as p
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> powers = [Component('P{}'.format(i), 1e-6, 2e-4) for i in (0,1)]
            >>> S = System()
            >>> S['E'] = [powers[0], powers[1]]
            >>> S[powers[0]] = S[powers[1]] = [motor]
            >>> S[motor] = 'S'
            >>> S.draw()
            >>> p.show()

        """
        nx.draw_graphviz(self._graph)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import print_function, absolute_import
import unittest2

from sympy import symbols, exp
from networkx import DiGraph, is_isomorphic

from fiabilipy import Component, Voter, System

class TestComponent(unittest2.TestCase):
    """ Test the Component class.
    """
    def setUp(self):
        """ Here we build the component we will use as test subjects
            We assume components have constant failure and maintainability rates
        """

        self.lambda_, self.mu = symbols('l, m',
                                        constant=True,
                                        positive=True,
                                        null=False
                                       )
        self.t = symbols('t', positive=True)
        self.component = Component('C', self.lambda_, self.mu)

    def test_reliability(self):
        """ Check the reliability is equals to the theorical one """
        self.assertEqual(exp(-self.lambda_ * self.t),
                         self.component.reliability(self.t))

    def test_maintainability(self):
        """ Check the maintainability is equals to the theorical one """
        self.assertEqual(1 - exp(-self.mu * self.t),
                         self.component.maintainability(self.t))

    def test_availability(self):
        """ Check the availability is equals to the theorical one """
        availability = self.lambda_ * \
                       exp(self.t*(-self.lambda_ - self.mu)) / \
                       (self.lambda_ + self.mu) + \
                       self.mu/(self.lambda_ + self.mu)

        self.assertEqual(availability,
                         self.component.availability(self.t))

class TestVoter(unittest2.TestCase):
    """ Test the Voter class.
    """
    pass

class TestSystem(unittest2.TestCase):
    """ Test the System class.
    """

    def setUp(self):
        """ Here we build some standard systems we will use as test subjects
        """
        systems = {'simple':System(),
                   'series-parallel':System(),
                   'parallel-series':System(),
                   'complex':System(),
                   'voter':System(),
                  }

        lambdas = {'alim': symbols('l_alim', positive=True, null=False),
                   'motor': symbols('l_motor', positive=True, null=False),
                  }
        mus = {'alim': symbols('m_alim', positive=True, null=False),
               'motor': symbols('m_motor', positive=True, null=False),
              }

        alim = [Component('Alim_A', lambda_=lambdas['alim'], mu=mus['alim']),
                Component('Alim_B', lambda_=lambdas['alim'], mu=mus['alim']),
                Component('Alim_C', lambda_=lambdas['alim'], mu=mus['alim']),
               ]
        motors = [
                Component('Motor_A', lambda_=lambdas['motor'], mu=mus['motor']),
                Component('Motor_B', lambda_=lambdas['motor'], mu=mus['motor']),
                 ]

        voter = Voter(alim[0], M=2, N=3)

        systems['simple']['E'] = alim[0]
        systems['simple'][alim[0]] = motors[0]
        systems['simple'][motors[0]] = 'S'

        systems['series-parallel']['E'] = [alim[0], alim[1]]
        systems['series-parallel'][alim[0]] = motors[0]
        systems['series-parallel'][alim[1]] = motors[1]
        systems['series-parallel'][motors[0]] = 'S'
        systems['series-parallel'][motors[1]] = 'S'

        systems['parallel-series']['E'] = [alim[0], alim[1]]
        systems['parallel-series'][alim[0]] = [motors[0], motors[1]]
        systems['parallel-series'][alim[1]] = [motors[0], motors[1]]
        systems['parallel-series'][motors[0]] = 'S'
        systems['parallel-series'][motors[1]] = 'S'

        systems['complex']['E'] = [alim[0], alim[1], alim[2]]
        systems['complex'][alim[0]] = motors[0]
        systems['complex'][alim[1]] = [motors[0], motors[1]]
        systems['complex'][alim[2]] = motors[1]
        systems['complex'][motors[0]] = 'S'
        systems['complex'][motors[1]] = 'S'

        systems['voter']['E'] = voter
        systems['voter'][voter] = [motors[0], motors[1]]
        systems['voter'][motors[0]] = 'S'
        systems['voter'][motors[1]] = 'S'

        self.systems = systems
        self.alim = alim
        self.motors = motors
        self.voter = voter
        self.lambdas = lambdas
        self.mus = mus

    def test_successpaths(self):
        """ Check that all success paths are correctly found.
        """
        paths = {
            'simple' : set([('E', self.alim[0], self.motors[0], 'S')]),
            'series-parallel': set([('E', self.alim[0], self.motors[0], 'S'),
                                    ('E', self.alim[1], self.motors[1], 'S')]),
            'parallel-series': set([('E', self.alim[0], self.motors[0], 'S'),
                                    ('E', self.alim[0], self.motors[1], 'S'),
                                    ('E', self.alim[1], self.motors[1], 'S'),
                                    ('E', self.alim[1], self.motors[0], 'S')]),
            'complex': set([('E', self.alim[0], self.motors[0], 'S'),
                            ('E', self.alim[1], self.motors[0], 'S'),
                            ('E', self.alim[1], self.motors[1], 'S'),
                            ('E', self.alim[2], self.motors[1], 'S')]),
            'voter': set([('E', self.voter, self.motors[0], 'S'),
                          ('E', self.voter, self.motors[1], 'S')]),
        }

        for (name, S) in self.systems.iteritems():
            for path in S.successpaths:
                paths[name].remove(tuple(path))
            self.assertEqual(paths[name], set())

    def test_minimalcuts(self):
        """ Check for minimal cut of orders 1 and 2.
        """
        cuts = {
            1: {
                 'simple': set([frozenset([self.alim[0]]),
                                frozenset([self.motors[0]])]),
                 'series-parallel': set(),
                 'parallel-series': set(),
                 'complex': set(),
                 'voter': set([frozenset([self.voter])]),
               },
            2: {
                 'simple': set([frozenset([self.alim[0]]),
                                frozenset([self.motors[0]])]),
                 'series-parallel': set([
                        frozenset([self.alim[1], self.alim[0]]),
                        frozenset([self.alim[1], self.motors[0]]),
                        frozenset([self.motors[1], self.alim[0]]),
                        frozenset([self.motors[1], self.motors[0]]),
                                        ]),
                 'parallel-series': set([
                        frozenset([self.alim[1], self.alim[0]]),
                        frozenset([self.motors[1], self.motors[0]])]),
                 'complex': set([frozenset([self.motors[0], self.motors[1]])]),
                 'voter': set([frozenset([self.voter]),
                               frozenset([self.motors[0], self.motors[1]])]),
               },
        }

        for order in [1, 2]:
            for (name, S) in self.systems.iteritems():
                for cut in S.minimalcuts(order):
                    cuts[order][name].remove(cut)
                self.assertEqual(cuts[order][name], set([]))

    def test_mttfvalues(self):
        r""" Check if the calculated MTTF values are correct.
             Testing MTTF values is interesting because there are computed by
             integration of reliability from 0 to \inf.
             So if the values of MTTF are correct, it means :
              - MTTF values are correct
              - Reliabitily value for any t are correct too.
             The drawback is that if this test fails, we don’t known which of
             MTTF property or reliability method is failing.
        """

        la = self.lambdas['alim']
        lm = self.lambdas['motor']
        mttf = {'simple': 1.0/(la + lm),
                'series-parallel': 3.0/(2*(la + lm)),
                'parallel-series':
                    1.0/(2*la + 2*lm) - 2.0/(2*la + lm) \
                    - 2.0/(la + 2*lm) + 4.0/(la + lm),
                'complex':
                    4.0/(la + lm) - 1.0/(2*lm + la) + 1.0/(2*lm + 3*la) \
                    - 1.0/(2*lm + 2*la) - 2.0/(2*la + lm),
                'voter':
                    6.0/(2*la + lm) - 3.0/(2*la + 2*lm) - 6.0/(3*la + lm) \
                    + 3.0/(3*la + 2*lm) + 2.0/(3*la + lm) - 1.0/(3*la + 2*lm)
               }

        for (name, values) in mttf.iteritems():
            diff = values - self.systems[name].mttf
            self.assertEqual(diff.simplify(), 0)

    def test_graphmanagement(self):
        """ Check if the constructing a system by its graph works as intended.
        """
        component = [Component('C%s' % i, 1e-3) for i in xrange(4)]
        system = System()

        #because 'E' must be the first inserted element
        with self.assertRaises(ValueError):
            system[component[0]] = 'S'

        #Assert the following constructions don’t fail.
        #from a list
        system['E'] = [component[0], component[1]]
        system[component[0]] = 'S'
        wanted = DiGraph({'E':[component[0], component[1]], component[0]:['S']})
        self.assertTrue(is_isomorphic(system._graph, wanted))

        del system[component[0]] #This component isn’t used anymore
        #from a single element
        system['E'] = component[1]
        system[component[1]] = 'S'
        wanted = DiGraph({'E':[component[1]], component[1]:'S'})
        self.assertTrue(is_isomorphic(system._graph, wanted))

    def test_cache(self):
        """ Perfom some tests on the cache
        """
        components = [Component('C{}'.format(i), 1e-3) for i in (0, 1, 2)]
        system = System()

        #      +-- C0 --+
        #      |        |
        # E ---|        +-- C2 -- S
        #      |        |
        #      +-- C1 --+

        system['E'] = [components[0], components[1]]
        system[components[0]] = components[2]
        system[components[1]] = components[2]
        system[components[2]] = 'S'

        self.assertAlmostEqual(system.mttf, 2000/3.)
        self.assertIn('mttf', system._cache) #The mttf is cached

        components[0].lambda_ = 0.05 #Let’s change the failure rate
        self.assertEqual(system._cache, dict()) #The cache is now empty
        self.assertAlmostEqual(system.mttf, 331750/663.)
        self.assertIn('mttf', system._cache) #The mttf is cached

        #now, check if it works with a shared component
        othersystem = System()
        othersystem['E'] = components[0]
        othersystem[components[0]] = 'S'

        self.assertAlmostEqual(othersystem.mttf, 20)
        components[0].lambda_ = 2e-4
        self.assertAlmostEqual(othersystem.mttf, 5000)
        self.assertAlmostEqual(system.mttf, 29000/33.)

if __name__ == '__main__':
    unittest2.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#Copyright (C) 2013 Chabot Simon, Sadaoui Akim

#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License along
#with this program; if not, write to the Free Software Foundation, Inc.,
#51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

r""" Component design module

This modude gives tools to design basic components and compute some metrics,
such as the reliability, the availability, the Mean-Time-To-Failure, and so on.

"""

from sympy import exp, Symbol, oo

__all__ = ['Component']

class Component(object):
    r""" Describe a component with a constant failure rate.

        This class is used to create all the components of a system.

        Attributes
        ----------
        name : str
            the name of the component. (It has to be a unique name for the whole
            system)
        lambda_ : float
            the constant failure rate of the component
        mu : float, optional
            the constant maintainability rate of the component
        initialy_avaible : boolean, optional
            whether the component is avaible at t=0 or not

        Examples
        --------
        >>> motor = Component('M', 1e-4, 3e-2)
        >>> motor.lambda_
        0.0001
    """

    def __init__(self, name, lambda_, mu=0, initialy_avaible=True):
        self.__dict__["_systems"] = set()
        self.lambda_ = lambda_
        self.mu = mu
        self.name = name
        self.initialy_avaible = initialy_avaible

    def __setattr__(self, name, value):
        for system in self._systems:
            system._cache = {}
        self.__dict__[name] = value

    def __repr__(self):
        return u'Component(%s)' % self.name

    def reliability(self, t):
        r""" Compute the reliability of the component at `t`

            This method compute the reliability of the component at `t`.

            Parameters
            ----------
            t : float or Symbol

            Returns
            -------
            out : float or symbolic expression
                The reliability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> t = Symbol('t', positive=True)
            >>> motor.reliability(t)
            exp(-0.0001*t)
            >>> motor.reliability(1000)
            0.904837418035960
        """
        return exp(-self.lambda_ * t)

    def maintainability(self, t):
        r""" Compute the maintainability of the component at `t`

            This method compute the maintainability of the component at `t`.

            Parameters
            ----------
            t : int or Symbol

            Returns
            -------
            out : float or symbolic expression
                The maintainability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> t = Symbol('t', positive=True)
            >>> motor.maintainability(t)
            -exp(-0.03*t) + 1.0
            >>> motor.maintainability(1000)
            0.999999999999906
        """
        return 1.0 - exp(-self.mu * t)

    def availability(self, t):
        r""" Compute the availability of the component at `t`

            This method compute the availability of the component at `t`.

            Parameters
            ----------
            t : int or Symbol

            Returns
            -------
            out : float or symbolic expression
                The availability calculated for the given `t`

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> t = Symbol('t', positive=True)
            >>> motor.availability(t)
            0.00332225913621263*exp(-0.0301*t) + 0.996677740863787
            >>> motor.availability(1000)
            0.996677740863788
        """
        if self.mu == self.lambda_ == 0:
            return 1
        a = self.mu / (self.mu + self.lambda_)
        if self.initialy_avaible:
            b = self.lambda_ / (self.mu + self.lambda_)
        else:
            b = - self.mu / (self.mu + self.lambda_)

        return a + b*exp(-(self.lambda_ + self.mu) * t)

    @property
    def mttf(self):
        r""" Compute the Mean-Time-To-Failure of the component

            The MTTF is defined as :
                :math:`MTTF = \int_{0}^{\infty} R(t)dt = \frac{1}{\lambda}`

            when the failure rate (:math:`\lambda` is constant)

            Returns
            -------
            out : float
                The component MTTF

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> motor.mttf
            10000.0
        """
        return 1.0/self.lambda_

    @property
    def mttr(self):
        r""" Compute the Mean-Time-To-Repair of the component

            The MTTR is defined as :
                :math:`MTTR = \int_{0}^{\infty} 1 - M(t)dt = \frac{1}{\mu}`

            when the failure rate (:math:`\mu` is constant)

            Returns
            -------
            out : float
                The component MTTR

            Examples
            --------
            >>> motor = Component('M', 1e-4, 3e-2)
            >>> motor.mttr
            33.333333333333336
        """
        return 1.0/self.mu
